#!/usr/bin/env python3

"""
    developer shell
    ~~~~~~~~~~~~~~~

    Just call this file, and the magic happens ;)

    This file is from: https://pypi.org/project/dev-shell/
    Source: https://github.com/jedie/dev-shell/blob/main/devshell.py

    :copyleft: 2021 by Jens Diemer
    :license: GNU GPL v3 or above
"""

import argparse
import hashlib
import signal
import subprocess
import sys
import venv
from pathlib import Path


try:
    import ensurepip  # noqa
except ModuleNotFoundError as err:
    print(err)
    print('-' * 100)
    print('Error: Pip not available!')
    print('Hint: "apt-get install python3-venv"\n')
    raise


assert sys.version_info >= (3, 7), 'Python version is too old!'


if sys.platform == 'win32':  # wtf
    # Files under Windows, e.g.: .../.venv/Scripts/python.exe
    BIN_NAME = 'Scripts'
    FILE_EXT = '.exe'
else:
    # Files under Linux/Mac and all other than Windows, e.g.: .../.venv/bin/python
    BIN_NAME = 'bin'
    FILE_EXT = ''

BASE_PATH = Path(__file__).parent
VENV_PATH = BASE_PATH / '.venv'
BIN_PATH = VENV_PATH / BIN_NAME
PYTHON_PATH = BIN_PATH / f'python{FILE_EXT}'
PIP_PATH = BIN_PATH / f'pip{FILE_EXT}'
POETRY_PATH = BIN_PATH / f'poetry{FILE_EXT}'

DEP_LOCK_PATH = BASE_PATH / 'poetry.lock'
DEP_HASH_PATH = VENV_PATH / '.dep_hash'

# script file defined in pyproject.toml as [tool.poetry.scripts]
# (Under Windows: ".exe" not added!)
PROJECT_SHELL_SCRIPT = BIN_PATH / 'devshell'


def get_dep_hash():
    """ Get SHA512 hash from poetry.lock content. """
    return hashlib.sha512(DEP_LOCK_PATH.read_bytes()).hexdigest()


def store_dep_hash():
    """ Generate /.venv/.dep_hash """
    DEP_HASH_PATH.write_text(get_dep_hash())


def venv_up2date():
    """ Is existing .venv is up-to-date? """
    if DEP_HASH_PATH.is_file():
        return DEP_HASH_PATH.read_text() == get_dep_hash()
    return False


def verbose_check_call(*popen_args):
    popen_args = [str(arg) for arg in popen_args]  # e.g.: Path() -> str for python 3.7
    print(f'\n+ {" ".join(popen_args)}\n')
    return subprocess.check_call(popen_args)


def noop_signal_handler(signal_num, frame):
    """
    Signal handler that does nothing: Used to ignore "Ctrl-C" signals
    """
    pass


def main(argv):
    assert DEP_LOCK_PATH.is_file(), f'File not found: "{DEP_LOCK_PATH}" !'

    if len(argv) == 2 and argv[1] in ('--update', '--help'):
        parser = argparse.ArgumentParser(
            prog=Path(__file__).name,
            description='Developer shell',
            epilog='...live long and prosper...'
        )
        parser.add_argument(
            '--update', default=False, action='store_true',
            help='Force create/upgrade virtual environment'
        )
        parser.add_argument(
            'command_args',
            nargs=argparse.ZERO_OR_MORE,
            help='arguments to pass to dev-setup shell/cli',
        )
        options = parser.parse_args(argv)
        force_update = options.update
        extra_args = argv[2:]
    else:
        force_update = False
        extra_args = argv[1:]

    # Create virtual env in ".../.venv/":
    if not PYTHON_PATH.is_file() or force_update:
        print('Create virtual env here:', VENV_PATH.absolute())
        builder = venv.EnvBuilder(symlinks=True, upgrade=True, with_pip=True)
        builder.create(env_dir=VENV_PATH)

    # install/update "pip" and "poetry":
    if not POETRY_PATH.is_file() or force_update:
        # Note: Under Windows pip.exe can't replace this own .exe file, so use the module way:
        verbose_check_call(PYTHON_PATH, '-m', 'pip', 'install', '-U', 'pip', 'setuptools')
        verbose_check_call(PIP_PATH, 'install', 'poetry')

    # install via poetry, if:
    #   1. .venv not exists
    #   2. "--update" used
    #   3. poetry.lock file was changed
    if not PROJECT_SHELL_SCRIPT.is_file() or force_update or not venv_up2date():
        verbose_check_call(POETRY_PATH, 'install')
        store_dep_hash()

    # The cmd2 shell should not abort on Ctrl-C => ignore "Interrupt from keyboard" signal:
    signal.signal(signal.SIGINT, noop_signal_handler)

    # Run project cmd shell via "setup.py" entrypoint:
    # (Call it via python, because Windows sucks calling the file direct)
    try:
        verbose_check_call(PYTHON_PATH, PROJECT_SHELL_SCRIPT, *extra_args)
    except subprocess.CalledProcessError as err:
        sys.exit(err.returncode)


if __name__ == '__main__':
    main(sys.argv)



import os
import sys
from pathlib import Path

import cmd2 as cmd2

from dev_shell.config import DevShellConfig
from dev_shell.utils.colorful import bright_blue, bright_yellow
from dev_shell.utils.subprocess_utils import argv2str


class DevShellBaseApp(cmd2.Cmd):
    """
    Base cmd2 App with some gimmicks ;)
    """

    def __init__(self, *args, config: DevShellConfig, **kwargs):
        kwargs.update(dict(
            persistent_history_file='.history',
            allow_cli_args=False
        ))
        super().__init__(*args, **kwargs)

        self.config = config

        self.intro = (
            f'\n\nDeveloper shell - {bright_yellow(config.package_path.name)} - v{config.version}\n'
        )
        self.prompt = f'\n({bright_blue(config.package_path.name)}) '

        # work-a-round: https://github.com/python-cmd2/cmd2/issues/1072
        args = sys.argv[1:]
        if args:
            # CLI usage => execute command and exit
            self.intro = None
            command = argv2str(args)
            self._startup_commands = [command]
        else:
            self._startup_commands = ['help']

        # Set aliases:
        self.aliases.update({
            'q': 'quit',
        })

        # Display Tracebacks on errors:
        self.debug = True

        self.update_path()

    def onecmd(self, *args, **kwargs):
        """
        1. Exit cmdloop if shell was used as CLI ;)
        2. Pass sys.exit() code
        """
        try:
            stop = super().onecmd(*args, **kwargs)
        except SystemExit as exit_code:
            # Pass the sys.exit() code
            # See also: https://github.com/python-cmd2/cmd2/pull/1076
            self.exit_code = exit_code
            stop = True

        return stop

    def update_path(self):
        """
        Add our .venv/bin/ directory into PATH at first position.
        """
        bin_path = str(Path(sys.executable).parent.absolute())
        env_path = os.environ.get('PATH', '')
        if not env_path.startswith(bin_path):
            os.environ['PATH'] = bin_path + os.pathsep + env_path


def run_cmd2_app(app: cmd2.Cmd) -> None:
    """
    Run a cmd2 App as CLI or shell.
    Handle exit code return value, too.
    """
    if len(sys.argv) > 1:
        # CLI usage => run only one command and then exit
        app.onecmd_plus_hooks(line=argv2str(sys.argv[1:]))
        exit_code = app.exit_code
    else:
        exit_code = app.cmdloop()

    sys.exit(exit_code)



import sys

from dev_shell.constants import BIN_PATH


if sys.platform == 'win32':
    VENV_PYTHON = BIN_PATH / 'python.exe'
    VENV_PIP = BIN_PATH / 'pip.exe'
    VENV_POETRY = BIN_PATH / 'poetry.exe'
    VENV_DEVSHELL = BIN_PATH / 'devshell'  # No ".exe" !
else:
    VENV_PYTHON = BIN_PATH / 'python'
    VENV_PIP = BIN_PATH / 'pip'
    VENV_POETRY = BIN_PATH / 'poetry'
    VENV_DEVSHELL = BIN_PATH / 'devshell'

DEVSHELL_CALL = f'{VENV_PYTHON} {VENV_DEVSHELL}'  # e.g.: ".venv/bin/python .venv/bin/devshell"



import filecmp
import shutil
import subprocess
import sys
from unittest import TestCase

from dev_shell.constants import BASE_PATH, BOOTSTRAP_SOURCE_FILE
from dev_shell.utils.assertion import assert_is_file


class SourceFileTestCase(TestCase):
    def test_source_file_is_up2date(self):
        own_bootstrap_file = BASE_PATH / 'devshell.py'
        assert_is_file(own_bootstrap_file)

        are_the_same = filecmp.cmp(own_bootstrap_file, BOOTSTRAP_SOURCE_FILE, shallow=False)
        if not are_the_same:
            shutil.copyfile(
                src=own_bootstrap_file,
                dst=BOOTSTRAP_SOURCE_FILE,
                follow_symlinks=False
            )
            raise AssertionError(f'Bootstrap source "{BOOTSTRAP_SOURCE_FILE}" updated!')

    def test_wrong_call(self):
        # The bootstrap script should only work if "poetry.lock" file exists in same path!
        p = subprocess.run(
            [sys.executable, str(BOOTSTRAP_SOURCE_FILE)],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            check=False
        )
        assert p.returncode == 1
        output = p.stdout.strip()
        assert 'File not found' in output
        if sys.platform == 'win32':
            assert output.endswith(r'\dev_shell\poetry.lock" !')
        else:
            assert output.endswith('/dev_shell/poetry.lock" !')



import subprocess
import sys
from pathlib import Path
from unittest.mock import patch

import pytest
from cmd2 import CommandResult

import dev_shell
from dev_shell.tests.fixtures import DevShellAppBaseTestCase
from dev_shell.tests.utils import call_mocked_subprocess
from dev_shell.utils.assertion import assert_is_file


OWN_DEV_SHELL_PATH = Path(dev_shell.__file__).parent.parent / 'devshell.py'
OWN_DEV_SHELL_PATH = OWN_DEV_SHELL_PATH.relative_to(Path().cwd())
assert_is_file(OWN_DEV_SHELL_PATH)


class DevShellAppTestCase(DevShellAppBaseTestCase):
    def test_help_raw(self):
        out = self.app.app_cmd('help')

        assert isinstance(out, CommandResult)
        assert 'Documented commands' in out.stdout

        assert 'Documented commands' in out.stdout

    def test_help_via_execute(self):
        stdout, stderr = self.execute('help')
        assert stderr == ''
        assert 'Documented commands' in stdout

    def test_do_pytest(self):
        with patch.object(subprocess, 'check_call') as check_call_mock:
            stdout, stderr = self.execute(command='pytest')

        assert stderr == ''
        check_call_mock.assert_called_once()
        popenargs = check_call_mock.call_args[0][0]
        command = popenargs[0]
        assert 'pytest' in command

        # The call will be printed:
        assert 'pytest' in stdout

    def test_do_update(self):
        origin_sys_argv = sys.argv.copy()
        try:
            sys.argv = ['TheDevScript.py']
            check_calls, (stdout, stderr) = call_mocked_subprocess(
                'check_call',
                self.execute,
                command='update'
            )

            assert stderr == ''

            assert len(check_calls) == 1
            if sys.platform == 'win32':
                assert check_calls[0].endswith(r'.venv\Scripts\poetry.exe update')
                assert '\n+ .venv\\Scripts\\poetry.exe update\n' in stdout
            else:
                assert check_calls[0].endswith('.venv/bin/poetry update')
                assert '\n+ .venv/bin/poetry update\n' in stdout

            assert '\nPlease restart "TheDevScript.py" !\n' in stdout
        finally:
            sys.argv = origin_sys_argv

    def test_do_linting(self):
        with patch.object(subprocess, 'check_call') as check_call_mock:
            stdout, stderr = self.execute(command='linting')

        print(stdout)
        print(stderr)

        assert stderr == ''

        # The call will be printed:
        if sys.platform == 'win32':
            assert '+ .venv\\Scripts\\darker.exe --diff --check\n' in stdout
            assert '+ .venv\\Scripts\\flake8.exe\n' in stdout
        else:
            assert '+ .venv/bin/darker --diff --check\n' in stdout
            assert '+ .venv/bin/flake8\n' in stdout

        check_call_mock.assert_called()

    def test_fix_code_style(self):
        with patch.object(subprocess, 'check_call') as check_call_mock:
            stdout, stderr = self.execute(command='fix_code_style')

        print(stdout)
        print(stderr)

        assert stderr == ''

        # The call will be printed:
        if sys.platform == 'win32':
            assert '+ .venv\\Scripts\\darker.exe\n' in stdout
        else:
            assert '+ .venv/bin/darker\n' in stdout

        check_call_mock.assert_called()

    def test_return_code(self):
        """
        If pytest failed, the cmd2 app should sys.exit() with >0 return code.
        Otherwise it's not useable in CI pipelines ;)

        This tests also the "run as CLI" implementation.
        """
        def call_pytest(*args):
            return subprocess.run(
                [sys.executable, str(OWN_DEV_SHELL_PATH), 'pytest'] + list(args),
                capture_output=True,
                text=True
            )

        # Provoke an error by trying to test a path that does not exist:
        p = call_pytest('/path/does/not/exists/')
        assert p.returncode > 0
        assert 'file or directory not found: /path/does/not/exists/' in p.stderr
        assert 'pytest /path/does/not/exists/' in p.stdout
        assert f'finished with exit code {p.returncode}' in p.stdout

        # Do something that normally should never fail:
        p = call_pytest('--version')
        assert p.stderr == f'pytest {pytest.__version__}\n'
        assert 'finished with exit code' not in p.stdout
        assert p.returncode == 0



import dataclasses
from pathlib import Path
from typing import Any


@dataclasses.dataclass
class DevShellConfig:
    package_module: Any = None  # The module for this shell

    # Set by post init from package_module:
    version: str = dataclasses.field(init=False)
    package_path: Path = dataclasses.field(init=False)
    base_path: Path = dataclasses.field(init=False)

    def __post_init__(self):
        assert self.package_module, 'Package module must be set!'

        assert hasattr(self.package_module, '__version__'), (
            f'{self.package_module!r} must have a __version__ attribute!'
        )
        assert hasattr(self.package_module, '__file__'), (
            f'{self.package_module!r} must have a __file__ attribute!'
        )

        self.version = self.package_module.__version__
        self.package_path = Path(self.package_module.__file__).parent
        self.base_path = self.package_path.parent



import os
import re
import shlex
import shutil
import subprocess
import sys
from pathlib import Path

from dev_shell.utils.assertion import assert_is_dir
from dev_shell.utils.colorful import blue, bright_blue, bright_white, bright_yellow, cyan, green, print_error


def argv2str(argv):
    """
    >>> argv2str(['foo', '--bar=123'])
    'foo --bar=123'
    """
    items = []
    for item in argv:
        if isinstance(item, Path):
            item = str(item)

        if re.match(r'^[-0-9a-zA-Z_.=]+$', item):
            items.append(item)
        else:
            items.append(shlex.quote(item))

    return ' '.join(items)


def make_absolute_path(path):
    """
    Resolve the path, but not for symlinks.
    e.g.: '.venv/bin/python' is a symlink to e.g.: '/usr/bin/python3.x'
    We would like to keep the origin symlink ;)
    """
    if not isinstance(path, Path):
        path = Path(path)

    if path.is_symlink():
        path = path.absolute()
    else:
        path = path.resolve(strict=True)

    return path


def make_relative_path(path, relative_to):
    """
    Makes {path} relative to {relative_to}, if possible.
    """
    path = make_absolute_path(path=path)
    relative_to = make_absolute_path(path=relative_to)

    try:
        is_relative_to = path.is_relative_to(relative_to)
    except AttributeError:  # is_relative_to() is new in Python 3.9
        try:
            path = path.relative_to(relative_to)
        except ValueError:
            # {path} doesn't start with {relative_to} -> do nothing
            pass
    else:
        if is_relative_to:
            path = path.relative_to(relative_to)

    return path


def _print_info(popenargs, *, cwd, kwargs):
    print()
    print('_' * 100)

    if len(popenargs) > 1:
        command, *args = popenargs
        args = argv2str(args)
    else:
        command = popenargs[0]
        args = ''

    command_path = make_relative_path(Path(command), relative_to=cwd)
    cwd = make_relative_path(cwd, relative_to=Path.cwd())

    command_name = command_path.name
    command_dir = command_path.parent

    info = ''

    if cwd.absolute() != Path.cwd().absolute():
        info = f'{bright_blue(str(cwd))}{bright_white("$")} '

    if command_dir and command_dir != Path.cwd():
        info += green(f'{command_dir}{os.sep}')

    if command_name:
        info += bright_yellow(command_name)

    if args:
        info += f' {blue(args)}'

    msg = f'+ {info}'

    verbose_kwargs = ', '.join(f'{k}={v!r}' for k, v in sorted(kwargs.items()))
    if verbose_kwargs:
        msg += f' (kwargs: {cyan(verbose_kwargs)})'

    print(f'{msg}\n', flush=True)


def prepare_popenargs(popenargs, cwd=None):
    if cwd is None:
        cwd = Path.cwd()
    else:
        assert_is_dir(cwd)

    command_path = Path(popenargs[0])
    if not command_path.is_file():
        # Lookup in current venv bin path first:
        bin_path = str(Path(sys.executable).parent.absolute())
        command = shutil.which(str(command_path), path=bin_path)
        if not command:
            # Search in PATH for this command that doesn't point to a existing file:
            command = shutil.which(str(command_path))
            if not command:
                raise FileNotFoundError(f'Command "{popenargs[0]}" not found in PATH!')

        command = make_absolute_path(path=command)

        # Replace command name with full path:
        popenargs = list(popenargs)
        popenargs[0] = command

    return popenargs, cwd


def verbose_check_call(
        *popenargs,
        verbose=True,
        cwd=None,
        extra_env=None,
        exit_on_error=False,
        **kwargs):
    """ 'verbose' version of subprocess.check_call() """

    popenargs, cwd = prepare_popenargs(popenargs, cwd=cwd)

    if verbose:
        _print_info(popenargs, cwd=cwd, kwargs=kwargs)

    env = os.environ.copy()
    if extra_env:
        env.update(extra_env)

    try:
        return subprocess.check_call(
            [str(part) for part in popenargs],  # e.g.: Path() instance -> str,
            universal_newlines=True,
            env=env,
            cwd=cwd,
            **kwargs
        )
    except subprocess.CalledProcessError as err:
        if exit_on_error:
            if verbose:
                print_error(f'Process "{popenargs[0]}" finished with exit code {err.returncode!r}')
            sys.exit(err.returncode)
        raise


def verbose_check_output(*popenargs, verbose=True, cwd=None, extra_env=None, **kwargs):
    """ 'verbose' version of subprocess.check_output() """

    popenargs, cwd = prepare_popenargs(popenargs, cwd=cwd)

    if verbose:
        _print_info(popenargs, cwd=cwd, kwargs=kwargs)

    env = os.environ.copy()
    if extra_env:
        env.update(extra_env)

    try:
        output = subprocess.check_output(
            [str(part) for part in popenargs],  # e.g.: Path() instance -> str,
            universal_newlines=True,
            env=env,
            cwd=cwd,
            stderr=subprocess.STDOUT,
            **kwargs
        )
    except subprocess.CalledProcessError as err:
        print('\n***ERROR:')
        print(err.output)
        raise
    return output



import io
from contextlib import redirect_stderr, redirect_stdout
from unittest import mock

from cmd2.ansi import strip_style


def call_mocked_subprocess(func_name, func, *args, catch_sys_exit=False, **kwargs):
    with mock.patch(f'subprocess.{func_name}') as cm:
        try:
            result = func(*args, **kwargs)
        except SystemExit:
            if not catch_sys_exit:
                raise
            result = None

        check_calls = []
        for args_list in cm.call_args_list:
            # TODO: Use "call.kwargs" and "call.args" if we drop Python 3.7 support!
            args, kwargs = args_list

            popenargs = args[0]
            if isinstance(popenargs, str):
                command_str = popenargs
            else:
                command_str = ' '.join(popenargs)

            cwd = kwargs.get('cwd')
            if cwd:
                command_str = f'{cwd}$ {command_str}'

            check_calls.append(command_str)

        return check_calls, result


class RedirectStdOutErr:
    """
    Buffer stdout + stderr with optional strip_style() call
    """

    def __init__(self):
        self._buffer = io.StringIO()
        self._output = None

    def __enter__(self):
        self._stdout_redirect = redirect_stdout(self._buffer).__enter__()
        self._stderr_redirect = redirect_stderr(self._buffer).__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._output = self._buffer.getvalue()
        finally:
            self._stdout_redirect.__exit__(exc_type, exc_val, exc_tb)
            self._stderr_redirect.__exit__(exc_type, exc_val, exc_tb)

    def get_output(self, remove_ansi=False):
        output = self._output
        if remove_ansi:
            output = strip_style(output)
        return output



from pathlib import Path


def assert_is_dir(path):
    """
    Check if given path is a directory
    """
    if not isinstance(path, Path):
        path = Path(path)

    if not path.is_dir():
        raise NotADirectoryError(f'Directory does not exists: "{path}"')


def assert_is_file(path):
    """
    Check if given path is a file
    """
    if not isinstance(path, Path):
        path = Path(path)

    assert_is_dir(path.parent)

    if not path.is_file():
        raise FileNotFoundError(f'File does not exists: "{path}"')



from unittest.case import TestCase

from cmd2 import CommandResult
from cmd2.ansi import strip_style
from cmd2_ext_test import ExternalTestMixin

from dev_shell.dev_shell_app import DevShellApp, get_devshell_app_kwargs


class CmdAppBaseTestCase(TestCase):
    """
    Base class for cmd2 app test cases
    """

    def get_app_instance(self):
        """
        :return: cmd2 app instance
        """
        raise NotImplementedError

    def setUp(self):
        super().setUp()
        self.app = self.get_app_instance()
        self.app.fixture_setup()

    def tearDown(self):
        super().tearDown()
        self.app.fixture_teardown()

    def execute(self, command, remove_colors=True):
        out = self.app.app_cmd(command)

        assert isinstance(out, CommandResult)
        if out.stdout is None:
            stdout = ''
        else:
            stdout = str(out.stdout)
            if remove_colors:
                stdout = strip_style(stdout)

        if out.stderr is None:
            stderr = ''
        else:
            stderr = str(out.stderr)
            if remove_colors:
                stderr = strip_style(stderr)

        return stdout, stderr


class DevShellAppTester(ExternalTestMixin, DevShellApp):
    pass


class DevShellAppBaseTestCase(CmdAppBaseTestCase):
    """
    Base class for dev-shell tests
    """

    def get_app_instance(self):
        # Init the test app with the same kwargs as the real app
        # see: dev_shell.cmd2app.devshell_cmdloop()
        app = DevShellAppTester(**get_devshell_app_kwargs())
        return app



import dev_shell
from dev_shell.base_cmd2_app import DevShellBaseApp, run_cmd2_app
from dev_shell.command_sets.dev_shell_commands import DevShellCommandSet
from dev_shell.config import DevShellConfig


class DevShellApp(DevShellBaseApp):
    """
    The "dev-shell" Cmd2 App
    """
    pass


def get_devshell_app_kwargs():
    """
    Generate the kwargs for the cmd2 App.
    (Separated because we needs the same kwargs in tests)
    """
    config = DevShellConfig(package_module=dev_shell)

    # initialize all CommandSet() with context:
    kwargs = dict(
        config=config
    )

    app_kwargs = dict(
        config=config,
        command_sets=[
            DevShellCommandSet(**kwargs),
        ]
    )
    return app_kwargs


def devshell_cmdloop():
    """
    Entry point to start the "dev-shell" cmd2 app.
    Used in: [tool.poetry.scripts]
    """
    app = DevShellApp(**get_devshell_app_kwargs())
    run_cmd2_app(app)  # Run a cmd2 App as CLI or shell



from cmd2 import CommandSet

from dev_shell.config import DevShellConfig


class DevShellBaseCommandSet(CommandSet):
    def __init__(self, config: DevShellConfig):
        super().__init__()

        self.config = config



import sys
from pathlib import Path

import cmd2

from dev_shell.command_sets import DevShellBaseCommandSet
from dev_shell.utils.colorful import bright_green
from dev_shell.utils.subprocess_utils import verbose_check_call


def run_linters(cwd=None):
    """
    Run code formatters and linter
    """
    verbose_check_call('darker', '--diff', '--check', cwd=cwd, exit_on_error=True)
    verbose_check_call('flake8', cwd=cwd, exit_on_error=True)


def run_fix_code_style(cwd=None):
    verbose_check_call('darker', cwd=cwd)


@cmd2.with_default_category('dev-shell commands')
class DevShellCommandSet(DevShellBaseCommandSet):
    """
    This command set may be used in external project, too.
    """

    def do_pytest(self, statement: cmd2.Statement):
        """
        Run dev-shell tests via pytest
        """
        verbose_check_call(
            'pytest',
            *statement.arg_list,
            cwd=self.config.base_path,
            exit_on_error=True
        )

    def do_update(self, statement: cmd2.Statement):
        """
        Call "poetry update" to update all dependencies in .venv
        """
        verbose_check_call(
            'poetry', 'update',
            *statement.arg_list,
            cwd=self.config.base_path,
            exit_on_error=True
        )
        script_name = Path(sys.argv[0]).name
        print(bright_green(f'\n\nPlease restart "{script_name}" !\n'))
        sys.exit(0)  # Stop cmd

    def do_linting(self, statement: cmd2.Statement):
        """
        Linting: Check code style with darker and flake8
        """
        verbose_check_call('darker', '--diff', '--check', cwd=self.config.base_path)
        verbose_check_call('flake8', cwd=self.config.base_path, exit_on_error=True)

    def do_fix_code_style(self, statement: cmd2.Statement):
        """
        Fix code style by running darker
        """
        run_fix_code_style(cwd=self.config.base_path)

    def do_list_venv_packages(self, statement: cmd2.Statement):
        """
        Just call "pip freeze" to list all installed venv packages
        """
        verbose_check_call(
            'pip', 'freeze',
            cwd=self.config.base_path
        )

    def do_publish(self, statement: cmd2.Statement):
        """
        Publish "dev-shell" to PyPi
        """
        # Maybe a project that use dev-shell doesn't use poetry-publish, too!
        # So import it here to make this dependency "optional"
        from poetry_publish.publish import poetry_publish  # noqa

        # Don't publish if test failed or code linting wrong:
        verbose_check_call(
            'pytest', '-x',
            cwd=self.config.base_path,
            exit_on_error=True,
        )
        run_linters(cwd=self.config.base_path)

        poetry_publish(
            package_root=self.config.base_path,
            version=self.config.version,
        )


# Maybe a project that use dev-shell doesn't use poetry-publish:
try:
    import poetry_publish  # noqa
except ModuleNotFoundError:
    # Remove the "publish" command
    del DevShellCommandSet.do_publish



"""
    Just some shortcuts to print colorful stuff
"""
import cmd2
from cmd2 import fg


def red(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.red, bold=False)


def green(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.green, bold=False)


def yellow(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.yellow, bold=False)


def blue(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.blue, bold=False)


def magenta(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.magenta, bold=False)


def cyan(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.cyan, bold=False)


def white(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.white, bold=False)


###################################################################################################


def bright_red(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_red, bold=False)


def bright_green(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_green, bold=False)


def bright_yellow(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_yellow, bold=False)


def bright_blue(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_blue, bold=False)


def bright_magenta(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_magenta, bold=False)


def bright_cyan(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_cyan, bold=False)


def bright_white(msg: str) -> str:
    return cmd2.ansi.style(msg, fg=fg.bright_white, bold=False)


###################################################################################################


def print_error(err: str) -> None:
    print(bright_red(f'\n{err}\n'))



"""
    Note: It's not needed to test the complete bootstrap of a new .venv
    Because the CI will automatically failed if the shell is not setup right ;)
"""
import io
import sys
from contextlib import redirect_stderr, redirect_stdout
from hashlib import sha512
from pathlib import Path
from unittest import TestCase, mock

import devshell
from dev_shell.constants import VENV_PATH
from dev_shell.tests.constants import DEVSHELL_CALL, VENV_DEVSHELL, VENV_PIP, VENV_POETRY, VENV_PYTHON
from dev_shell.tests.utils import call_mocked_subprocess
from dev_shell.utils.assertion import assert_is_dir, assert_is_file


def call_devsetup_main(*args, catch_sys_exit=False):
    out = io.StringIO()
    err = io.StringIO()
    with redirect_stdout(out), redirect_stderr(err):
        check_calls, result = call_mocked_subprocess(
            'check_call',
            devshell.main,
            args,
            catch_sys_exit=catch_sys_exit
        )

    if sys.platform == 'win32':
        # FIXME: e.g.: https://github.com/jedie/dev-shell/runs/2272270967
        check_calls = [entry.replace("'", '').replace('"', '') for entry in check_calls]

    return check_calls, out.getvalue(), err.getvalue()


class BootstrapTestCase(TestCase):
    def test_up2date_bootstrap(self):
        # .dep_hash is up-to-date -> no "poetry install" call:
        check_calls, stdout, stderr = call_devsetup_main()
        assert stderr == ''
        assert f'{DEVSHELL_CALL}\n' in stdout
        assert check_calls == [DEVSHELL_CALL]
        assert 'poetry' not in stdout

    def test_poetry_install_call(self):
        base_path = Path(devshell.__file__).parent
        assert_is_dir(base_path / 'dev_shell')

        DEP_HASH_PATH = base_path / '.venv' / '.dep_hash'
        assert_is_file(DEP_HASH_PATH)

        DEP_LOCK_PATH = base_path / 'poetry.lock'

        current_hash = sha512(DEP_LOCK_PATH.read_bytes()).hexdigest()

        assert DEP_HASH_PATH.read_text() == current_hash

        # Change the hash -> "poetry install" should be called:

        try:
            DEP_HASH_PATH.write_text('this is not the hash')

            check_calls, stdout, stderr = call_devsetup_main()
            assert stderr == ''
            assert f'{VENV_POETRY} install' in stdout
            assert f'{VENV_PYTHON} {VENV_DEVSHELL}\n' in stdout
            assert check_calls == [
                f'{VENV_POETRY} install',  # <<< install called?
                DEVSHELL_CALL
            ]
        finally:
            DEP_HASH_PATH.write_text(current_hash)

    def test_force_update(self):
        # EnvBuilder.create() will ran into "SameFileError" because we call the create()
        # from the same venv to update the same venv ;)
        # So we mock this call here.
        with mock.patch('venv.EnvBuilder.create') as create_mock:
            check_calls, stdout, stderr = call_devsetup_main('devshell.py', '--update')

        assert stderr == ''
        assert f'{VENV_POETRY} install' in stdout
        assert f'{DEVSHELL_CALL}\n' in stdout
        assert check_calls == [
            f'{VENV_PYTHON} -m pip install -U pip setuptools',
            f'{VENV_PIP} install poetry',
            f'{VENV_POETRY} install',
            DEVSHELL_CALL
        ]
        create_mock.assert_called_once_with(env_dir=VENV_PATH)

    def test_help(self):
        check_calls, stdout, stderr = call_devsetup_main(
            'devshell.py', '--help',
            catch_sys_exit=True
        )
        assert stderr == ''
        assert 'usage: devshell.py [-h] [--update] [command_args ' in stdout
        assert stdout.endswith('...live long and prosper...\n')
        assert check_calls == []

    def test_pass_help(self):
        """
        A "--help" argument should be pass to a cmd2 command
        """
        check_calls, stdout, stderr = call_devsetup_main(
            'devshell.py', 'not_existing_command', '--help',
            catch_sys_exit=True
        )
        assert stderr == ''
        assert 'not_existing_command --help' in stdout
        assert check_calls == [
            f'{DEVSHELL_CALL} not_existing_command --help'
        ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dev_shell', 'dev_shell.command_sets', 'dev_shell.tests', 'dev_shell.utils']

package_data = \
{'': ['*']}

install_requires = \
['cmd2']

extras_require = \
{':sys_platform == "darwin"': ['gnureadline']}

entry_points = \
{'console_scripts': ['devshell = dev_shell.dev_shell_app:devshell_cmdloop']}

setup_kwargs = {
    'name': 'dev-shell',
    'version': '0.3.0',
    'description': 'Devloper shell for easy startup...',
    'long_description': '# A "dev-shell" for Python projects ;)\n\n[![pytest](https://github.com/jedie/dev-shell/actions/workflows/test.yml/badge.svg?branch=main)](https://github.com/jedie/dev-shell/actions?query=branch%3Amain)\n[![codecov](https://codecov.io/gh/jedie/dev-shell/branch/main/graph/badge.svg)](https://codecov.io/gh/jedie/dev-shell)\n\nThis small project is intended to improve the start-up for collaborators.\n\nThe idea is to make the project setup as simple as possible. Just clone the sources and start a script and you\'re done ;)\n\nRun Tests? Just start the script and call the "run test command".\n\nThe "dev-shell" is the base to create a CLI and a shell. It also\n\nIt also shows how to make a project bootstrap as simply as possible, e.g.:\n\n```bash\n~$ git clone https://github.com/jedie/dev-shell.git\n~$ cd dev-shell\n~/dev-shell$ ./devshell.py pytest\n```\n\n\n## How it works\n\nFirst start of the Python script [./devshell.py](https://github.com/jedie/dev-shell/blob/main/devshell.py) will bootstrap:\n\n* Generate a Python virtual environment (in short: `venv`)\n* Install poetry\n* Install project dependencies and the project himself\n\nThe output on first bootstrap start looks like:\n\n```bash\n~/dev-shell$ ./devshell.py\nCreate venv here: ~/dev-shell/.venv\nCollecting pip\n...\nSuccessfully installed pip-21.0.1\nCollecting poetry\n...\nInstalling dependencies from lock file\n\nPackage operations: 31 installs, 1 update, 0 removals\n\n...\n\nInstalling the current project: dev-shell (0.0.1alpha0)\n\n\n+ .venv/bin/python .venv/bin/devshell\n\n\nDeveloper shell - dev_shell - v0.2.0\n\n\nDocumented commands (use \'help -v\' for verbose/\'help <topic>\' for details):\n\ndev-shell commands\n==================\nfix_code_style  linting  list_venv_packages  publish  pytest  update\n\n...\n\n(dev_shell) quit\n~/dev-shell$\n```\n\nThe first bootstrap start takes a few seconds. Each later startup detects the existing virtualenv and is very fast:\n\n```bash\n~/dev-shell$ ./devshell.py\n\nDeveloper shell - dev_shell - v0.2.0\n\n(dev_shell) help\n```\n\nInfo: The `.venv` will be automatically updated via `poetry install` call if the `poetry.lock` file has been changed.\n\nA call with `--update` will force to call some create/update steps, e.g.:\n\n```bash\n~/dev-shell$ ./devshell.py --update\n```\n\nYou can also just delete `/.venv/` and start `devshell.py` again ;)\n\n(Using `--update` is not to be confused with the call of "update" command.)\n\n\n## compatibility\n\n| dev-shell version | OS                      | Python version |\n|-------------------|-------------------------|----------------|\n| >=v0.0.1          | Linux + MacOS + Windows | 3.9, 3.8, 3.7  |\n\nSee also github test configuration: [.github/workflows/test.yml](https://github.com/jedie/dev-shell/blob/main/.github/workflows/test.yml)\n\n## History\n\n* [*dev*](https://github.com/jedie/dev-shell/compare/v0.3.0...main)\n  * TBC\n* [v0.3.0 - 2022-01-30](https://github.com/jedie/dev-shell/compare/v0.2.4...v0.3.0)\n  * Remove "flynt" form linting/fix code style\n* [v0.2.4 - 2022-01-30](https://github.com/jedie/dev-shell/compare/v0.2.3...v0.2.4)\n  * Update requirements\n  * Use darker as code formatter and pytest-darker for linting\n* [v0.2.3 - 2021-11-15](https://github.com/jedie/dev-shell/compare/v0.2.2...v0.2.3)\n  * Update requirements\n  * [Flynt arguments can be changes via CommandSet](https://github.com/jedie/dev-shell/issues/29)\n* [v0.2.2 - 2021-04-13](https://github.com/jedie/dev-shell/compare/v0.2.1...v0.2.2)\n  * Include bootstrap file, to it\'s possible to use it in external projects, too.\n* [v0.2.1 - 2021-04-12](https://github.com/jedie/dev-shell/compare/v0.2.0...v0.2.1)\n  * Handle if "poetry-publish" is not installed, so a project that used "dev-shell" must not install it.\n* [v0.2.0 - 2021-04-11](https://github.com/jedie/dev-shell/compare/v0.1.0...v0.2.0)\n  * Rename: "dev-shell.py => devshell.py" because of better autocomplete\n  * Add `DevShellConfig.base_path` and use it in own commands like, `pytest`, `linting` etc. (So they are usable in external project, too.)\n  * recognize "--update" and "--help" arguments better in `./devshell.py` calls.\n  * Update `setuptools` on `.venv` creation, too.\n  * Fix Bugs/tests under Windows\n* [v0.1.0 - 2021-03-22](https://github.com/jedie/dev-shell/compare/v0.0.2...v0.1.0)\n  * Fix CI usage: Exit with correct return code if tests failed\n  * Better "run as CLI" implementation via new `run_cmd2_app()`\n  * Bugfix errors that only occur on Windows.\n  * Simplify `devshell.py` boot script and fix raise error if `ensurepip` missing\n* [v0.0.2 - 2021-03-19](https://github.com/jedie/dev-shell/compare/v0.0.1...v0.0.2)\n  * refactor colorful shortcuts\n  * display subprocess calls with separated colors\n* [v0.0.1 - 2021-03-19](https://github.com/jedie/dev-shell/compare/ad5dca...v0.0.1)\n  * first "useable" version\n\n## Project links\n\n* Github: https://github.com/jedie/dev-shell/\n* PyPi: https://pypi.org/project/dev-shell/\n',
    'author': 'Jens Diemer',
    'author_email': 'python@jensdiemer.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jedie/dev-shell',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0.0',
}


setup(**setup_kwargs)



import os
import platform
import sys
from pathlib import Path
from unittest import TestCase

from dev_shell.constants import BASE_PATH, BIN_PATH, VENV_PATH
from dev_shell.tests.constants import VENV_PYTHON
from dev_shell.tests.utils import RedirectStdOutErr, call_mocked_subprocess
from dev_shell.utils.subprocess_utils import (
    _print_info,
    argv2str,
    make_relative_path,
    prepare_popenargs,
    verbose_check_call,
    verbose_check_output,
)


class UtilsTestCase(TestCase):
    def test_argv2str(self):
        result = argv2str([Path('/foo/bar'), '--foo'])
        if sys.platform == 'win32':
            assert result == "'\\foo\\bar' --foo"
        else:
            assert result == '/foo/bar --foo'

    def assert_make_relative_path(self, path, relative_to, result):
        p = make_relative_path(path=Path(path), relative_to=Path(relative_to))
        assert p == Path(result)

    def test_basic(self):
        if sys.platform == 'win32':
            result = r'.venv/Scripts'
        else:
            result = '.venv/bin'

        self.assert_make_relative_path(
            path=VENV_PYTHON.parent,
            relative_to=VENV_PYTHON.parent.parent.parent,
            result=result
        )


class SubprocessUtilsTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.origin_cwd = Path.cwd()
        os.chdir(BASE_PATH)

    def tearDown(self):
        super().tearDown()
        os.chdir(self.origin_cwd)

    def test_prepare_popenargs(self):
        popenargs, cwd = prepare_popenargs(
            popenargs=[BASE_PATH / 'devshell.py'],
            cwd=None
        )
        assert popenargs == [BASE_PATH / 'devshell.py']
        assert cwd == BASE_PATH

        popenargs, cwd = prepare_popenargs(
            popenargs=['python', '--version'],
            cwd=None
        )
        assert popenargs == [VENV_PYTHON, '--version']
        assert cwd == BASE_PATH

        popenargs, cwd = prepare_popenargs(
            popenargs=['python', '--version'],
            cwd=Path.cwd()
        )
        assert popenargs == [VENV_PYTHON, '--version']
        assert cwd == BASE_PATH

        popenargs, cwd = prepare_popenargs(
            popenargs=['python', '--version'],
            cwd=BIN_PATH
        )
        assert popenargs == [VENV_PYTHON, '--version']
        assert cwd == BIN_PATH

        popenargs, cwd = prepare_popenargs(
            popenargs=['python', '--version'],
            cwd=BIN_PATH.parent  # .../dev-shell/.venv/
        )
        assert popenargs == [VENV_PYTHON, '--version']
        assert cwd == BIN_PATH.parent

        popenargs, cwd = prepare_popenargs(
            popenargs=['python', '--version'],
            cwd=Path(__file__).parent
        )
        assert popenargs == [VENV_PYTHON, '--version']
        assert cwd == Path(__file__).parent

        with self.assertRaises(NotADirectoryError) as cm:
            prepare_popenargs(
                popenargs=['does not exists'],
                cwd='/does/not/exists/'
            )
        if sys.platform == 'win32':
            assert cm.exception.args[0] == r'Directory does not exists: "\does\not\exists"'
        else:
            assert cm.exception.args[0] == 'Directory does not exists: "/does/not/exists"'

        with self.assertRaises(FileNotFoundError) as cm:
            prepare_popenargs(popenargs=['does not exists'])
        assert cm.exception.args[0] == 'Command "does not exists" not found in PATH!'

    def test_print_info(self):
        def call_print_info(*args, **kwargs):
            with RedirectStdOutErr() as redirector:
                _print_info(*args, **kwargs)

            output = redirector.get_output(remove_ansi=True)
            output = output.strip()
            output = output.lstrip('\n_')
            return output

        output = call_print_info(
            popenargs=[VENV_PYTHON, '--version'],
            cwd=Path.cwd(),
            kwargs={}
        )
        if sys.platform == 'win32':
            assert output == r'+ .venv\Scripts\python.exe --version'
        else:
            assert output == '+ .venv/bin/python --version'

        output = call_print_info(
            popenargs=[str(VENV_PYTHON), '--version'],
            cwd=Path(BIN_PATH / '..').resolve(),  # .../dev-shell/.venv/,
            kwargs={}
        )
        if sys.platform == 'win32':
            assert output == r'+ .venv$ Scripts\python.exe --version'
        else:
            assert output == '+ .venv$ bin/python --version'

    def test_verbose_check_call(self):
        with RedirectStdOutErr() as redirector:
            check_calls, result = call_mocked_subprocess(
                'check_call',
                verbose_check_call,
                str(VENV_PYTHON), '--version',
                cwd=Path(BIN_PATH / '..').resolve(),  # .../dev-shell/.venv/,
                shell=False
            )
            assert check_calls == [f'{VENV_PATH}$ {VENV_PYTHON} --version']

        output = redirector.get_output(remove_ansi=True)
        output = output.strip()
        output = output.lstrip('\n_')
        if sys.platform == 'win32':
            assert output == r'+ .venv$ Scripts\python.exe --version (kwargs: shell=False)'
        else:
            assert output == '+ .venv$ bin/python --version (kwargs: shell=False)'

    def test_check_output_mocked(self):
        with RedirectStdOutErr() as redirector:
            check_calls, result = call_mocked_subprocess(
                'check_output',
                verbose_check_output,
                VENV_PYTHON, '--version',
                cwd=Path(BIN_PATH / '..').resolve(),  # .../dev-shell/.venv/
                shell=False
            )
            assert check_calls == [f'{VENV_PATH}$ {VENV_PYTHON} --version']

        output = redirector.get_output(remove_ansi=True)
        output = output.strip()
        output = output.lstrip('\n_')
        if sys.platform == 'win32':
            assert output == r'+ .venv$ Scripts\python.exe --version (kwargs: shell=False)'
        else:
            assert output == '+ .venv$ bin/python --version (kwargs: shell=False)'

    def test_check_output(self):
        with RedirectStdOutErr() as redirector:
            output = verbose_check_output(
                VENV_PYTHON, '--version',
                cwd=Path(BIN_PATH / '..').resolve(),  # .../dev-shell/.venv/
                shell=False
            )
            assert output == f'Python {platform.python_version()}\n'

        output = redirector.get_output(remove_ansi=True)
        output = output.strip()
        output = output.lstrip('\n_')
        if sys.platform == 'win32':
            assert output == r'+ .venv$ Scripts\python.exe --version (kwargs: shell=False)'
        else:
            assert output == '+ .venv$ bin/python --version (kwargs: shell=False)'



import sys
from pathlib import Path


BASE_PATH = Path(__file__).parent.parent.resolve()
assert Path(BASE_PATH / 'dev_shell').is_dir(), f'Path wrong: {BASE_PATH}'


if sys.platform == 'win32':  # wtf
    BIN_NAME = 'Scripts'
else:
    BIN_NAME = 'bin'

VENV_PATH = BASE_PATH / '.venv'

BIN_PATH = VENV_PATH / BIN_NAME
assert BIN_PATH.is_dir()


# The source file path for external projects:
BOOTSTRAP_SOURCE_FILE = BASE_PATH / 'dev_shell' / 'bootstrap-source.py'
